/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.util.Spliterator;
import java.util.function.DoubleConsumer;
import java.util.stream.DoubleStream;

/**
 * Immutable array of double primitives.
 *
 * @author Leon van Zantvoort
 */
@Immutable
public interface DoubleArray extends Array<Double> {

    /** Empty array. */
    static final DoubleArray EMPTY = unsafeValueOf();

    /**
     * Creates an immutable copy of the specified {@code array}. If the specified {@code array} is mutable, changes to
     * the mutable array are not reflected by this immutable copy.
     *
     * <p>The specified {@code array} can be any array of primitive or object numbers. Reconstructing a copy of this
     * array may involve rounding or truncation as implemented by {@link Number#doubleValue}. The specified {@code array}
     * must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code MUTABLE} and {@code DYNAMIC}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return immutable array of double values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    static DoubleArray copyOf(@Nonnull Array<? extends Number> array) {
        return Arrays.newArray(UnsafeArray.doubleReadAccess(array), array.offset(), array.length(),
                array.characteristics(), array.hasCharacteristics(MUTABLE) && UnsafeArray.hasDirectAccess(array, Double.TYPE));
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return immutable array of double values.
     */
    static DoubleArray copyOf(@Nonnull double... array) {
        return copyOf(array, 0, array.length);
    }

    /**
     * <p>Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.</p>
     *
     * <p>Note that the backing array will have the same length as {@code array}.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return immutable array of double values.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static DoubleArray copyOf(@Nonnull double[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, true);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array. It is therefore advised not to modify {@code array} after this immutable array is
     * created, as this would violate this array's immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return immutable array that is backed by the specified {@code array}.
     */
    static DoubleArray unsafeValueOf(@Nonnull double... array) {
        return unsafeValueOf(array, 0, array.length);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array. It is therefore advised not to modify {@code array} after this immutable array is
     * created, as this would violate this array's immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return immutable array that is backed by the specified {@code array}.
     */
    static DoubleArray unsafeValueOf(@Nonnull double[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, false);
    }

    /**
     * Returns the double value found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return double value found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    double getDouble(int index);

    /**
     * Returns {@code true} if this array contains the specified value.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified value.
     */
    boolean containsDouble(double value);

    /**
     * Returns the index of the first occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    int indexOfDouble(double value);

    /**
     * Returns the index of the last occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified value is found.</p>
     *
     * @param value value whose highest index in this array is to be returned.
     * @return highest index of value in array or {@code -1} if value is not found.
     */
    int lastIndexOfDouble(double value);

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within array.
     *
     * @param consumer consumer that is called for each value within array's range.
     */
    void forEachDouble(@Nonnull DoubleConsumer consumer);

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    double[] toArray();

    /**
     * Returns an array containing all of the double values of {@code this} array. If the double values fit in the
     * specified array, it is returned therein. Otherwise, a new double array is allocated.
     *
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    double[] toArray(@Nonnull double[] array);

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    DoubleArray offset(int offset);    // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}
     */
    @Override
    DoubleArray length(int length);    // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    DoubleArray subArray(int fromIndex, int toIndex);  // More specific return type.

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    ArrayIterator.OfDouble iterator(); // More specific return type.

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    ArrayIterator.OfDouble iterator(int index);   // More specific return type.

    /**
     * Creates a {@code Spliterator} over the elements in this array.
     *
     * <p>The {@code Spliterator} reports:
     * <ul>
     *     <li>{@code Spliterator.SIZED};</li>
     *     <li>{@code Spliterator.SUBSIZED};</li>
     *     <li>{@code Spliterator.ORDERED};</li>
     *     <li>{@code Spliterator.IMMUTABLE} if {@code hasCharacteristics(MUTABLE)} returns {@code false};</li>
     *     <li>{@code Spliterator.NONNULL} if {@code hasCharacteristics(NONNULL)} returns {@code true};</li>
     *     <li>{@code Spliterator.SORTED} if {@code hasCharacteristics(SORTED)} returns {@code true};</li>
     *     <li>{@code Spliterator.DISTINCT} if {@code hasCharacteristics(DISTINCT)} returns {@code true}.</li>
     * </ul>
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     * return Arrays.spliterator(array);
     * </pre>
     *
     * @return a {@code Spliterator} over the elements in this array.
     */
    @Override
    default Spliterator.OfDouble spliterator() {   // More specific return type.
        return Arrays.spliterator(this);
    }

    /**
     * Returns a sequential primitive int stream for this array.
     *
     * @return sequential primitive int stream.
     */
    default DoubleStream doubleStream() {
        return Arrays.doubleStream(this);
    }

    /**
     * Returns a parallel primitive int stream for this array.
     *
     * @return parallel primitive int stream.
     */
    default DoubleStream doubleParallelStream() {
        return Arrays.doubleParallelStream(this);
    }
}
