/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;

/**
 * Dynamic-size array of char primitives.
 *
 * @author Leon van Zantvoort
 */
public interface DynamicCharArray extends MutableCharArray, DynamicArray<Character> {

    /**
     * Creates an empty dynamic array with an initial capacity of ten.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code DYNAMIC}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @return new dynamic array instance.
     */
    static DynamicCharArray newInstance() {
        return newInstance(10);
    }

    /**
     * Creates an empty array with the specified initial capacity.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code DYNAMIC}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param  initialCapacity  the initial capacity of the array.
     * @throws IllegalArgumentException if the specified initial capacity is negative.
     * @return dynamic array.
     */
    static DynamicCharArray newInstance(int initialCapacity) {
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal capacity: " + initialCapacity);
        }
        return valueOf(new char[initialCapacity], 0, 0);
    }

    /**
     * Creates dynamic-size, mutable copy of the specified {@code array}. If the specified {@code array} is mutable,
     * changes to the mutable array are not reflected by this dynamic copy.
     *
     * <p>The specified {@code array} can be an array of {@code char} primitives or {@code Character} objects. The
     * specified {@code array} must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code DYNAMIC};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code MUTABLE} and {@code DYNAMIC}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return dynamic-size mutable array of char values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static DynamicCharArray copyOf(@Nonnull Array<Character> array) {
        return Arrays.newDynamicArray(UnsafeArray.charReadAccess(array), array.offset(), array.length(),
                array.characteristics(), UnsafeArray.hasDirectAccess(array, Character.TYPE));
    }

    /**
     * Creates a dynamic array of the specified {@code array}. The specified {@code array} is used as backing array,
     * but is replaced by a larger array when more capacity is needed.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code DYNAMIC};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return a dynamic array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static DynamicCharArray valueOf(@Nonnull char... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates a dynamic array of the specified {@code array}. The specified {@code array} is used as backing array,
     * but is replaced by a larger array when more capacity is needed.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code DYNAMIC};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return a dynamic array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static DynamicCharArray valueOf(@Nonnull char[] array, int offset, int length) {
        return Arrays.newDynamicArray(array, offset, length, 0, false);
    }

    /**
     * Appends the specified char value to the end of this array's range. Length of {@code array} is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param value char value to to be added.
     */
    void addChar(char value);

    /**
     * Inserts the specified char value at the specified {@code index}, moving elements at and after {@code index} to the
     * right. Length of {@code this} array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param value char value to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    void addChar(int index, char value);

    /**
     * Removes element at specified {@code index}, moving all elements after {@code index} to the left. Length of
     * {@code this} array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} is reported.
     * </ul>
     *
     * @param index removal index, relative to the offset of the array.
     * @return element that is removed.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    char charRemove(int index);

    /**
     * Removes the first occurrence of the specified value from this array if it is present. If this array does not
     * contain the value, it is unchanged.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics to lookup the element to be removed, otherwise all elements are traversed until the specified
     * value is found.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} is reported.
     * </ul>
     *
     * @param value value to be removed from this array.
     * @return {@code true} if this array contained the specified value.
     */
    boolean removeChar(char value);

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    DynamicCharArray offset(int offset);     // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}
     */
    @Override
    DynamicCharArray length(int length);     // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    DynamicCharArray subArray(int fromIndex, int toIndex);   // More specific return type.
}
