/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Objects;
import java.util.function.Consumer;

/**
 * Immutable view of Java Collection Framework's List.
 *
 * @author Leon van Zantvoort
 */
@Immutable
class ListArrayImpl<E> implements ObjectArray<E>, UnsafeArray<E>, Serializable {

    /** backing list. */
    transient List<E> elements;

    /** Start index of range. */
    final int offset;

    /** Length of range. */
    int length;

    /** Array characteristics. */
    int characteristics;

    /** Comparator that is used for sorting array. */
    Comparator<? super E> comparator;

    /**
     * Package visible constructor, to be used by static factory methods only.
     *
     * @param list backing list.
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *                         as backing array.
     */
    ListArrayImpl(@Nonnull List<E> list, int offset, int length, int characteristics, @Nullable Comparator<? super E> comparator, boolean copy) {
        Objects.requireNonNull(list);
        if (offset < 0 || length < 0 || length > list.size() - offset) {
            throw new IndexOutOfBoundsException();
        }
        if (copy) {
            this.elements = new ArrayList<>(list);
            this.offset = 0;
        } else {
            this.elements = list;
            this.offset = offset;
        }
        this.length = length;
        this.characteristics = filterCharacteristics(characteristics) | (length > 0 ? 0 : NONNULL | SORTED | DISTINCT);
        this.comparator = comparator;
    }

    /**
     * Checks if the given index is in range.  If not, throws an appropriate
     * runtime exception.  This method does *not* check if the index is
     * negative: It is always used immediately prior to an array access,
     * which throws an IndexOutOfBoundsException if index is negative.
     */
    final void rangeCheck(int index) {
        Arrays.rangeCheck(length, index);
    }

    /**
     * Checks if the given boundary is within this array's boundaries.
     */
    final void subArrayCheck(int offset, int length) {
        Arrays.subArrayCheck(this.offset, this.length, offset, length);
    }

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    @Override
    public Class<? super E> getComponentType() {
        return Object.class;
    }

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @Override
    public @CheckForNull Comparator<? super E> getComparator() {
        return comparator;
    }

    /**
     * Access to the backing array not supported for this implementation. This method throws an
     * {@code UnsupportedOperationException}.
     *
     * @return reference to the backing array.
     * @throws UnsupportedOperationException is always thrown by this method.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public Object[] array() {
        throw new UnsupportedOperationException();
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    protected int filterCharacteristics(int characteristics) {
        return characteristics & ~(MUTABLE | DYNAMIC);
    }

    /**
     * Returns the characteristics set for this array. The result is represented as ORed values from
     * {@link #NONNULL}, {@link #MUTABLE}, {@link #DYNAMIC}, {@link #SORTED}, {@link #DISTINCT}.
     * Modification of the array contents may result in updated characteristics.
     *
     * @return the characteristics set for this array.
     */
    @Override
    public final int characteristics() {
        return characteristics;
    }

    /**
     * <p>Returns the offset relative to the backing array.</p>
     *
     * @return offset relative to the backing array.
     */
    @Override
    public final int offset() {
        return offset;
    }

    /**
     * <p>Returns the number of elements of this array. All index related methods must respect the boundary represented
     * by this value.</p>
     *
     * <p>Note that for mutable arrays, the backing array - as returned by {@code MutableArray.array()} -
     * may be larger.</p>
     *
     * @return length of array.
     */
    @Override
    public final int length() {
        return length;
    }

    /**
     * Returns {@code true} if this array contains no elements.
     *
     * @return {@code true} if this array contains no elements.
     */
    @Override
    public final boolean isEmpty() {
        return length == 0;
    }

    /**
     * Returns the object found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @SuppressWarnings("unchecked")
    @Override
    public @Nullable E get(int index) {
        rangeCheck(index);
        return elements.get(offset + index);
    }

    /**
     * Returns the object found at {@code [offset + index]} of the backing array. This method allows for fast
     * failure detection.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if null reference is found at index.
     */
    @SuppressWarnings("unchecked")
    @Override
    public @Nonnull E getNonnull(int index) {
        rangeCheck(index);
        E value = elements.get(offset + index);
        Objects.requireNonNull(value, "Assertion error, nonnull reference expected.");
        return value;
    }

    /**
     * Returns {@code true} if this array contains the specified object. This method supports boxed primitives. This
     * means that if this array represents a primitive array, this method returns {@code true} if this array contains
     * the unboxed version of specified object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param object object whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified object.
     * @throws NullPointerException if this array is sorted and object is null.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public boolean contains(Object object) {
        if (hasCharacteristics(SORTED)) {
            return Arrays.unsafeBinarySearch(this, (E) object, getComparator()) >= 0;
        } else {
            int toIndex = offset + length;
            for (int index = offset; index < toIndex; index++) {
                if (Objects.equals(elements.get(index), object)) {
                    return true;
                }
            }
            return false;
        }
    }

    /**
     * Returns {@code true} if this array contains all of the elements in the specified {@code array}. Primitives and
     * their object counterparts are regarded equal in this respect.
     *
     * <p>The default implementation performs a binary search if this array is sorted, otherwise all elements are
     * traversed until the specified value is found.</p>
     *
     * @param array array to be checked for containment in this array.
     * @return {@code true} if this array contains all of the elements in the specified array.
     * @throws NullPointerException if the specified array is null or if array is sorted and the specified contains
     * null elements.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with key.
     */
    @Override
    public boolean containsAll(Array<?> array) {
        for (Object o : array) {
            if (!contains(o)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns the index of the first occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with key.
     */
    @SuppressWarnings("unchecked")
    @Override
    public int indexOf(Object object) {
        if (hasCharacteristics(SORTED)) {
            Comparator<? super E> comparator = getComparator();
            int index = Arrays.unsafeBinarySearch(this, (E) object, comparator);
            if (index >= 0) {
                return Arrays.unsafeBinarySearchFirst(this, index, comparator);
            }
            return -1;
        }
        int toIndex = offset + length;
        ListIterator<E> it = elements.listIterator(offset);
        for (int index = offset; index < toIndex; index++) {
            if (Objects.equals(it.next(), object)) {
                return index;
            }
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified object is found.</p>
     *
     * @param object object whose highest index in this array is to be returned.
     * @return highest index of object in array or {@code -1} if value is not found.
     * @throws NullPointerException if this array is sorted and the specified array contains null elements.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with key.
     */
    @SuppressWarnings("unchecked")
    @Override
    public int lastIndexOf(Object object) {
        if (hasCharacteristics(SORTED)) {
            Comparator<? super E> comparator = getComparator();
            int index = Arrays.unsafeBinarySearch(this, (E) object, comparator);
            if (index >= 0) {
                return Arrays.unsafeBinarySearchLast(this, index, comparator);
            }
            return -1;
        }
        int fromIndex = offset;
        int toIndex = offset + length;
        ListIterator<E> it = elements.listIterator(toIndex);
        for (int index = toIndex -1; index >= fromIndex; index--) {
            if (Objects.equals(it.previous(), object)) {
                return index;
            }
        }
        return -1;
    }

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    public Object[] toArray() {
        return elements.subList(offset, offset + length).toArray();
    }

    /**
     * Returns an array containing all of the reference of {@code this} array. If the references fit in the
     * specified array, it is returned therein. Otherwise, a new array of the same object type is allocated.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T> T[] toArray(@Nonnull T[] array) {
        Objects.requireNonNull(array);
        //noinspection SuspiciousToArrayCall
        return elements.subList(offset, offset + length).toArray(array);
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within array.
     *
     * @param consumer consumer that is called for each value within array's range.
     */
    @Override
    public void forEach(@Nonnull Consumer<? super E> consumer) {
        Objects.requireNonNull(consumer);
        elements.subList(offset, offset + length).forEach(consumer);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public ObjectArray<E> offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new ListArrayImpl<>(elements, newOffset, newLength, characteristics, comparator, false);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}
     */
    @Override
    public ObjectArray<E> length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new ListArrayImpl<>(elements, offset, length, characteristics, comparator, false);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    @SuppressWarnings("unchecked")
    public ObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new ListArrayImpl<>(elements, newOffset, newLength, characteristics, comparator, false);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        final ListIterator<E> iterator;

        ArrayIteratorImpl(ListIterator<E> iterator) {
            this.iterator = iterator;
        }

        @Override
        public boolean hasNext() {
            return iterator.hasNext();
        }

        @Override
        public E next() {
            return iterator.next();
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return iterator.hasPrevious();
        }

        @Override
        public E previous() {
            return iterator.previous();
        }

        @Override
        public int nextIndex() {
            return iterator.nextIndex();
        }

        @Override
        public int previousIndex() {
            return iterator.previousIndex();
        }

        @Override
        public void set(E reference) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void add(E reference) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator());
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator(index));
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public String toString() {
        int iMax = offset + length - 1;
        if (iMax - offset == -1) {
            return "[]";
        }
        StringBuilder b = new StringBuilder();
        b.append('[');
        for (Iterator<E> it = iterator(); it.hasNext();) {
            b.append(it.next());
            if (it.hasNext()) {
                return b.append(']').toString();
            }
            b.append(", ");
        }
        return b.toString();
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public int hashCode() {
        int result = 1;
        //noinspection ForLoopReplaceableByForEach
        for (Iterator<E> it = iterator(); it.hasNext();) {
            Object element = it.next();
            result = 31 * result + (element == null ? 0 : element.hashCode());
        }
        return result;
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public boolean equals(Object obj) {
        return obj == this || obj instanceof Array && Arrays.equals(this, (Array) obj);
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = -6316889894958073614L;

        transient int offset;
        final int length;
        final int characteristics;
        final Comparator<? super E> comparator;
        transient List<E> elements;

        SerializationProxy(ListArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            s.writeObject(new ArrayList<>(elements.subList(offset, offset + length)));
        }

        @SuppressWarnings("unchecked")
        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (List<E>) s.readObject();
        }

        @SuppressWarnings("unchecked")
        private Object readResolve() {
            return new ListArrayImpl<>(elements, 0, length, characteristics, comparator, false);
        }
    }
}