/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.Objects;
import java.util.function.UnaryOperator;

/**
 * Base interface for mutable arrays.
 *
 * @param <E> object type.
 * @author Leon van Zantvoort
 */
public interface MutableArray<E> extends Array<E> {

    /**
     * Returns the backing array (optional operation). All elements of {@code this} mutable array and parents of this
     * mutable array are stored in the returned object. Modifications to the backing array are reflected by {@code this}
     * mutable array and it's parents.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing class.</p>
     *
     * <p>Default implementation throws UnsupportedOperationException.</p>
     *
     * @return backing array.
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    default Object array() {
        throw new UnsupportedOperationException("Backing array not exposed by implementation.");
    }

    /**
     * Replaces element at specified {@code index} by specified object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index index relative to the offset of the array.
     * @param object object to replace existing value.
     * @return object that is replaced.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if specified {@code object} is null and array does not permit null values.
     */
    @CheckForNull E set(int index, E object);

    /**
     * Replaces all elements of this array by the elements of the specified array. Note that the length of {@code array}
     * needs to be equal to the length of this array, throws an {@code IllegalArgumentException} otherwise.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED};</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws IllegalArgumentException if length of specified array is not equal to length of this array.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     */
    void setAll(@Nonnull Array<? extends E> array);

    /**
     * Performs the specified {@code ArrayAction} without affecting the array's set of characteristics. Ownership of
     * array characteristics control is moved to the caller by means of the {@code ArrayContext} object. Errors or
     * runtime exceptions thrown by the action are relayed to the caller.
     *
     * <p>The following example demonstrates how a {@code null} element is added to {@code array}, while maintaining
     * array's sort characteristics.
     * <pre>
     *{@literal array.sort();
     * boolean modified = array.doAction(context -> {
     *     array.add(null);
     *     context.removeCharacteristics(NULLABLE);
     *     return true;
     * }}</pre>
     *
     * @param <T> return type.
     * @param action the action to be performed.
     * @return the value returned by the action's {@code run} method.
     * @throws IllegalStateException if action is called when array does not own control of characteristics.
     * @exception NullPointerException if the action is {@code null}.
     */
    <T> T doAction(@Nonnull ArrayAction<T> action);

    /**
     * Replaces each element of this array with the result of applying the operator to that element. Errors or runtime
     * exceptions thrown by the operator are relayed to the caller.
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     *{@literal final ArrayIterator<E> it = array.iterator();
     * while (it.hasNext()) {
     *   it.set(operator.apply(it.next()));
     * }
     * }</pre>
     *
     * @param operator the operator to apply to each element
     * @throws NullPointerException if the specified operator is null.
     */
    default void replaceAll(@Nonnull UnaryOperator<E> operator) {
        Objects.requireNonNull(operator);
        final ArrayIterator<E> it = iterator();
        while (it.hasNext()) {
            it.set(operator.apply(it.next()));
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    MutableArray<E> offset(int offset);

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}
     */
    @Override
    MutableArray<E> length(int length);

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    MutableArray<E> subArray(int fromIndex, int toIndex);

    /**
     * Sorts this array according to natural ordering.
     */
    void sort();

    /**
     * Sorts this array according to natural ordering. The underlying implementation utilizes multiple threads
     * to reduce sorting time. The algorithm may require a working space equal to the size of the array.
     */
    void parallelSort();
}
