/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.util.Comparator;

/**
 * Fixed-size mutable array of object references.
 *
 * @author Leon van Zantvoort
 */
public interface MutableObjectArray<E> extends ObjectArray<E>, MutableArray<E> {

    /**
     * Creates a fixed-size, mutable copy of the specified {@code array}. If the specified {@code array} is mutable,
     * changes to the mutable array are not reflected by this mutable copy.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code MUTABLE} and {@code DYNAMIC}.</li>
     * </ul>
     *
     * @param <E> object type.
     * @param array array to be copied.
     * @return fixed-size mutable array of objects.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    static <E> MutableObjectArray<E> copyOf(@Nonnull Array<E> array) {
        return Arrays.newMutableArray((E[]) UnsafeArray.readAccess(array), array.offset(), array.length(),
                array.characteristics(), array.getComparator(), UnsafeArray.hasDirectAccess(array, Object.class));
    }

    /**
     * Creates a mutable array of the specified {@code array}. The specified {@code array} is
     * used as backing array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param <E> object type.
     * @return a mutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SafeVarargs
    static <E> MutableObjectArray<E> valueOf(@Nonnull E... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates a mutable array of the specified {@code array}. The specified {@code array} is
     * used as backing array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE};</li>
     *   <li>{@code NONNULL} if the specified {@code length} is zero;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param <E> object type.
     * @return a mutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static <E> MutableObjectArray<E> valueOf(@Nonnull E[] array, int offset, int length) {
        return Arrays.newMutableArray(array, offset, length, 0, null, false);
    }

    /**
     * Sorts this array's range using the supplied {@code Comparator} to compare elements.
     *
     * The default implementation is equivalent to, for this {@code array}:
     * <pre>
     * Arrays.sort(array, c);
     * </pre>
     *
     * @param comparator the {@code Comparator} used to compare list elements.
     *          A {@code null} value indicates that the elements'
     *          {@linkplain Comparable natural ordering} should be used
     * @throws ClassCastException if the list contains elements that are not
     *         <i>mutually comparable</i> using the specified comparator
     * @throws IllegalArgumentException
     *         (<a href="Collection.html#optional-restrictions">optional</a>)
     *         if the comparator is found to violate the {@link java.util.Comparator}
     *         contract
     * @throws NullPointerException if the specified comparator is null.
     */
    void sort(Comparator<? super E> comparator);

    /**
     * Sorts this array's range using the supplied {@code Comparator} to compare elements. The underlying implementation
     * utilizes multiple threads to reduce sorting time. The algorithm requires a working space equal to the size of the
     * to be sorted range.
     *
     * The default implementation is equivalent to, for this {@code array}:
     * <pre>
     * Arrays.parallelSort(array, c);
     * </pre>
     *
     * @param comparator the {@code Comparator} used to compare array elements.
     *          A {@code null} value indicates that the elements'
     *          {@linkplain Comparable natural ordering} should be used
     * @throws ClassCastException if the list contains elements that are not
     *         <i>mutually comparable</i> using the specified comparator
     * @throws IllegalArgumentException
     *         (<a href="Collection.html#optional-restrictions">optional</a>)
     *         if the comparator is found to violate the {@link java.util.Comparator}
     *         contract
     * @throws NullPointerException if the specified comparator is null.
     */
    void parallelSort(Comparator<? super E> comparator);

    /**
     * Returns the backing array (optional operation). All elements of {@code this} mutable array and parents of this
     * mutable array are stored in the returned object. Modifications to the backing array are reflected by {@code this}
     * mutable array and it's parents.
     *
     * <p>Default implementation throws UnsupportedOperationException.</p>
     *
     * @return backing array.
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    Object[] array();   // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array. The sub array has the same mutability properties as {@code this} array.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    MutableObjectArray<E> offset(int offset);   // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * <p>A defensive copy of the backing array is created for immutable implementations. Mutable implementations
     * share the same backing array.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}
     */
    @Override
    MutableObjectArray<E> length(int length);   // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    MutableObjectArray<E> subArray(int fromIndex, int toIndex); // More specific return type.
}
