/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Iterator;

/**
 * Arrays implementing this interface provide access to the backing array. This interface allows utility classes
 * such as {@link Arrays} to perform fast read operations on arrays.
 *
 * @author Leon van Zantvoort
 */
interface UnsafeArray<E> extends Array<E> {

    /**
     * Returns {@code true} if the specified {@code array} exposes the backing array, {@code false} otherwise.
     *
     * @param array array to be inspected.
     * @param componentType component type to test for.
     * @return {@code true} if the specified array exposes the backing array, {@code false} otherwise.
     */
    static boolean hasDirectAccess(Array<?> array, Class<?> componentType) {
        return array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess() && componentType.isAssignableFrom(array.getComponentType());
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable boolean[] booleanWriteAccess(BooleanArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (boolean[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable byte[] byteWriteAccess(ByteArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (byte[]) ((UnsafeArray) array).array();
        }
        return null;
    }


    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable char[] charWriteAccess(CharArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (char[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable short[] shortWriteAccess(ShortArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (short[]) ((UnsafeArray) array).array();
        }
        return array.toArray();
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable int[] intWriteAccess(IntArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (int[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable long[] longWriteAccess(LongArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (long[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable float[] floatWriteAccess(FloatArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (float[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    static @Nullable double[] doubleWriteAccess(DoubleArray array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (double[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, {@code null} otherwise.
     *
     * @param array array to obtain backing array from.
     * @return backing array or {@code null}.
     */
    @SuppressWarnings("unchecked")
    static @Nullable Object[] writeAccess(ObjectArray<?> array) {
        if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
            return (Object[]) ((UnsafeArray) array).array();
        }
        return null;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull boolean[] booleanReadAccess(Array<? extends Boolean> array) {
        if (array.getComponentType() == Boolean.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (boolean[]) ((UnsafeArray) array).array();
            }
            return (boolean[]) array.toArray();
        }
        boolean[] a = new boolean[array.length()];
        int length = array.length();
        Iterator<? extends Boolean> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next(); // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull byte[] byteReadAccess(Array<? extends Number> array) {
        if (array.getComponentType() == Byte.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (byte[]) ((UnsafeArray) array).array();
            }
            return (byte[]) array.toArray();
        }
        byte[] a = new byte[array.length()];
        int length = array.length();
        Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next().byteValue(); // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull char[] charReadAccess(Array<? extends Character> array) {
        if (array.getComponentType() == Character.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (char[]) ((UnsafeArray) array).array();
            }
            return (char[]) array.toArray();
        }
        char[] a = new char[array.length()];
        int length = array.length();
        Iterator<? extends Character> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next(); // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull short[] shortReadAccess(Array<? extends Number> array) {
        if (array.getComponentType() == Short.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (short[]) ((UnsafeArray) array).array();
            }
            return (short[]) array.toArray();
        }
        short[] a = new short[array.length()];
        int length = array.length();
        Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next().shortValue();    // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull int[] intReadAccess(Array<? extends Number> array) {
        if (array.getComponentType() == Integer.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (int[]) ((UnsafeArray) array).array();
            }
            return (int[]) array.toArray();
        }
        int[] a = new int[array.length()];
        int length = array.length();
        Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next().intValue();  // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access and specified {@code array}'s backing array
     * component type is primitive, a copy of the backing array is returned otherwise.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull long[] longReadAccess(Array<? extends Number> array) {
        if (array.getComponentType() == Long.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (long[]) ((UnsafeArray) array).array();
            }
            return (long[]) array.toArray();
        }
        long[] a = new long[array.length()];
        int length = array.length();
        Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next().longValue(); // Throws NullPointerException if element is null.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull float[] floatReadAccess(Array<? extends Number> array) {
        if (array.getComponentType() == Float.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (float[]) ((UnsafeArray) array).array();
            }
            return (float[]) array.toArray();
        }
        float[] a = new float[array.length()];
        int length = array.length();
        Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next().floatValue(); // Throws NullPointerException if element is null.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull double[] doubleReadAccess(Array<? extends Number> array) {
        if (array.getComponentType() == Double.TYPE) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (double[]) ((UnsafeArray) array).array();
            }
            return (double[]) array.toArray();
        }
        double[] a = new double[array.length()];
        int length = array.length();
        Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next().doubleValue();   // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns backing array if specified {@code array} has direct access, a copy of the backing array is returned
     * otherwise. This copy is obtained through the {@code }Array.toArray()} method.
     *
     * @param array array to obtain native array from.
     * @return backing array or copy of backing array.
     */
    @SuppressWarnings("unchecked")
    static @Nonnull Object[] readAccess(Array<?> array) {
        if (!array.getComponentType().isPrimitive()) {
            if (array instanceof UnsafeArray && ((UnsafeArray) array).hasDirectAccess()) {
                return (Object[]) ((UnsafeArray) array).array();
            }
            return (Object[]) array.toArray();
        }
        Object[] a = new Object[array.length()];
        int length = array.length();
        Iterator<?> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            a[index] = iterator.next();
        }
        return a;
    }

    /**
     * Returns {@code true} if this array exposes the backing array, {@code false} otherwise.
     *
     * @return {@code true} if this array exposes the backing array, {@code false} otherwise.
     */
    default boolean hasDirectAccess() {
        return false;
    }

    /**
     * Returns the backing array. All elements of {@code this} mutable array and parents of this
     * mutable array are stored in the returned object. Modifications to the backing array are reflected by {@code this}
     * mutable array and it's parents.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing class.</p>
     *
     * @return backing array.
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    default Object array() {
        if (hasDirectAccess()) {
            throw new AssertionError("UnsafeArray implementations must implement this method.");
        } else {
            throw new UnsupportedOperationException("Backing array not exposed by implementation.");
        }
    }
}
